(* Reste a ajouter un mot-cle symbol a la syntaxe commune pour resultat des 
   if *)

#open"syntaxe_commune";;
#open"syntaxe_scheme";;

type 'a option =
  Vide
| Objet of 'a
;;

let motcles = [|
  "if";"or";"cond";"case";"define";"let";"let*";"letrec";
  "begin";"do";"lambda";"quote";"set!"
|];;

(* Pour les applications sans arguments,  *)
(* on insere un argument artificiel       *)
(* Cela se voit dans la fonction abstrait *)
(* et dans la fonction applique plus loin *)

let rec abstrait e = function
    Atome Nil                            -> Abs ("",e)
  | Cons (Atome (Symbole id), Atome Nil) -> Abs (id, e)
  | Cons (Atome (Symbole id), l) -> Abs (id, abstrait e l)
  | _                            -> failwith "parametres mal construits"
;;

let apparait id e = app_rec e
  where rec app_rec = function
    App (e,e1)    ->
         (app_rec e) or (app_rec e1)
  | Abs (id',e)   ->
         if id = id' then false else app_rec e
  | Var id'       -> id = id'
  | If (e0,e1,e2) -> (app_rec e0) or (app_rec e1) or (app_rec e2)
  | Set (id',e)   -> id = id'
  | LetIn (idl,el,e)    ->
         (it_list (fun b e -> (app_rec e) or b) false el) or
         (if mem id idl then false else (app_rec e))
  | LetRecIn (idl,el,e) ->
         (it_list (fun b e -> (app_rec e) or b) false el) or
         (if mem id idl then false else (app_rec e))
  | _             -> false
;;

let rec trad_expr_scheme = function
   Cons (Atome (Symbole s),e) -> begin
     match (s,e) with
       ("if",Cons(e1,Cons(e2,Cons(e3,Atome Nil))))
           -> If (trad_expr_scheme e1,trad_expr_scheme e2,trad_expr_scheme e3)
     | ("if",Cons(e1,Cons(e2,Atome Nil)))
           -> If (trad_expr_scheme e1,trad_expr_scheme e2,Cst (Bool false))
     | ("define",e)
           -> define e Vide
     | ("set!",Cons(Atome(Symbole s),Cons(e2,Atome Nil)))
           -> Set (s,trad_expr_scheme e2)
     | ("set!",Cons(Atome(Symbole s),Atome Nil))
           -> Set (s,Cst Indefini)
     | ("let",Cons(e1,e2))
           -> let (noms,vals) = decomp e1 in
         LetIn (noms,vals,sequence e2)
     | ("letrec",Cons(e1,e2))
           -> let (noms,vals) = decomp e1 in
         LetRecIn (noms,vals,sequence e2)
     | ("lambda",Cons(e1,e2))
           -> abstrait (sequence e2) e1
     | ("begin",e)
           -> let rec f = function 
                  Cons(e1,Atome Nil) -> trad_expr_scheme e1
                | Cons(Cons(Atome(Symbole "define"),e),e2)        ->
                      define e (Objet (f e2))
                | Cons(e1,e2)        ->
                      LetIn([nouveau_nom ()],[trad_expr_scheme e1],(f e2))
                | _                  -> failwith "begin mal construit"
              in
                (f e)
    |  _
          -> applique (Var s) e
    end
 | Cons (e1,e2) -> applique (trad_expr_scheme e1) e2
 | Atome at     -> begin
     match at with
       True       -> Cst (Bool true)
     | Nil        -> Cst (Bool false)
     | Symbole id -> Var id
     | SString s  -> Cst (String s)
     | SInt i     -> Cst (Int i)
     | SFloat f   -> Cst (Float f)
     | SChar c    -> Cst (Char c)
     end

and applique f = function
    Atome Nil   -> App (f, Cst Unit)
  | Atome _     -> failwith "arguments mal construits"
  | Cons (e, Atome Nil) -> App (f,trad_expr_scheme e)
  | Cons (_, Atome _) -> failwith "arguments mal construits"
  | Cons (e, l) -> applique (App (f,trad_expr_scheme e)) l

and decomp = function
   Atome Nil -> ([],[])
 | Cons (Cons (Atome (Symbole s), Cons (e, Atome Nil)), reste) ->
   let (ids,vals) = decomp reste in
   (s::ids,(trad_expr_scheme e)::vals)
 | _         -> failwith "liaisons mal construites"

and sequence = function 
   Cons(e1,Atome Nil)                        ->
      trad_expr_scheme e1
 | Cons(Cons(Atome(Symbole "define"),e),e2)  ->
      define e (Objet (sequence e2))
 | Cons(e1,e2)                               ->
      LetIn([nouveau_nom ()],[trad_expr_scheme e1],(sequence e2))
 | _                  -> failwith "sequence mal construite"
  
and define e suite =
  let (id,trad_e) = match e with
     Cons(Atome(Symbole id),Atome Nil)
         -> (id,Cst Indefini)
   | Cons(_,Atome _)
         -> failwith "Expression incorrecte"
   | Cons(Atome(Symbole id), e1)
         -> (id,sequence e1)
   | Cons(Cons(Atome(Symbole id),e1), e2)
         -> (id,(abstrait (sequence e2) e1))
   | _   -> failwith "Nombre d'arguments incorrect"
  in
  let corps = match suite with
     Vide -> Cst (Nom id)
   | Objet trad_e' -> trad_e'
  in
    if (apparait id trad_e) then
      LetRecIn([id],[trad_e],corps)
    else
      LetIn([id],[trad_e],corps)
;;

let ajoute_init_env corps = 
  LetIn (["+";"*";"-";"/";"=";"display"],
         [Abs ("",Abs ("",Prim Plus));
          Abs ("",Abs ("",Prim Fois));
          Abs ("",Abs ("",Prim Moins));
          Abs ("",Abs ("",Prim Divise));
          Abs ("",Abs ("",Prim Egal));
          Abs ("",Prim Affiche)],
         corps)
;;

let rec internalise = function
    [] -> Atome Nil
  | a::reste -> Cons (a, internalise reste)
;;

let encapsule l =
  if l=[] then Atome Nil
  else Cons (Atome (Symbole "begin"), internalise l)
;;

let traduit_scheme l =
  ajoute_init_env (trad_expr_scheme (encapsule l))
;;

